#!/usr/bin/perl

#NAME
#   PFOR v0.965, January 2011. ( Author: Qingfa Wu, email: wuqf@ustc.edu.cn)

#SYNOPSIS
#   Viroids discovery by clustering short DNA sequences based on Kmer extension

#LICENSE
#   PFOR Copyright (c) 2011 University of California, Riverside. All rights Reserved.

#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.


use Getopt::Std;
use vars qw( %opt );
BEGIN { $| = 1 }

### Get command line options ###

&getopt('fokslnxDU', \%opt);

### If insufficient options, print usage###

if(! ($opt{'f'}) ) 
{
   print "\nUsage: $0\n\n";
   print "-f <file.fasta> input reads\n";  
   print "-o <output_file> sequences of viroid candidates (default: input_file.contigs)\n";
   print "-k <number> initial assembling k-mer (default = 17)\n"; 
   print "-s <number> minimal length of reads for assembling (default = 18)\n";
   print "-l <number> maximal length of reads for assembling (default = 40)\n";
   print "-n <number> the mininal number of reads for extension (default = 1)\n";
   print "-x <number> the range of test kmers (i.e., 17+x) (default: +0)\n";
   print "-D <number> the mininal length of candidate sequence (default = 100)\n";
   print "-U <number> the maximal length of candidate sequence (default = 800)\n\n";
}

### Set global parameters ###

my $inputfile  = $opt{'f'};
my $outputfile = "$inputfile".'.viroid';

my $kmer = 17;
my $minL = 18; 
my $maxL = 40;

my $minE = 1;
my $minAL =100;
my $maxAL =800;
my $range = 0;

##
if ($opt{'s'}) { $minL = $opt{'s'};} 
if ($opt{'l'}) { $maxL = $opt{'l'};}
if ($opt{'n'}) { $minE = $opt{'n'};}
if ($opt{'U'}) { $maxAL = $opt{'U'};}
if ($opt{'D'}) { $minAL = $opt{'D'};}
if ($opt{'k'}) { $kmer = $opt{'k'};}
if ($opt{'o'}) { $outputfile = $opt{'o'};}
if ($opt{'x'}) { $range = $opt{'x'};}


### test file ###

undef my $inputset;

#reading reads from input file
if(-e $inputfile) { 
  $inputset = &readINFILE($inputfile,$minL,$maxL);	
}
else { 
  die "\nPlease input the small RNA file (fasta format)!\n\n"; 
}


########################################
#test distinct kmers to filter reads and assembling the remaining reads.

for(my $testKmer=$kmer; $testKmer <= ($kmer+$range); $testKmer++)
{
  #removing non-overlapping reads and terminal small reads. 
  #The process would be repeated iteratively until no TSR is found and
  #all of the small RNAs remaining in the pool are true ISRs. 

  $inputset = filterINPUTset($inputset, $testKmer);
  
  #keep the filtered data
  keepFILTERdata($inputset,$testKmer);

  #assembling the remainingg reads 
  &testAssembleDistinctKmer( $inputset, $testKmer);
}




#################################################################
sub testAssembleDistinctKmer
{
   my ($ysinputset, $ykmer) = @_;

   #seeking all overlapping pairs
   my  $match = matchPAIRseek( $ysinputset, $ykmer);
   

   print "start assembling reads using k-mer of $ykmer\n";
   my $assNUM=0;

   open(OUT, ">$outputfile.$ykmer"); 
   foreach my $reads ( keys %$ysinputset )
   {  
     unless($reads=~/[ATCG]/ && length($reads) > $ykmer  ) { next; }

     $assNUM++; 
	 print "." if( $assNUM%100 == 0 );

     my $viroidseq = &AssemblyReads($match,$ysinputset,$reads,$ykmer,$minE,$minAL,$maxAL);
     if(length($viroidseq) >= $minAL && length($viroidseq) <= $maxAL ) { print OUT ">Contig-$assNUM\n$viroidseq\n" ; }   

   }
   close(OUT);
   print "\n";

   return 1;
	
}
#####################################################
#the following script is to seek all overlapping pairs.
####

sub  matchPAIRseek
{
   my ($putsetin, $ykmer) = @_;
   
   my $setbiny = &creatHASH($putsetin,$ykmer);

   my $ymatch;
   my $pcount=0;
   
   print "Pair set search using k-mer of $ykmer:\n";

   foreach my $reads (keys %$putsetin ) 
    {  
      my $pairset = &seekMATCHB($reads, $setbiny, $ykmer); 

      foreach my $pairs (keys %$pairset) 
       {
         my ($aread,$bread)=split(/,/,$pairs);
         $ymatch->{$aread}{$bread}++; 
	     $pcount++; 
		 print "." if ($pcount%100000==0);
       }
    }
   print "\n";

   return $ymatch;
}

#####################################################


sub AssemblyReads 
{
	my ($imatch,$iset,$iseq,$ikmer,$iminE,$imiAL,$imxAL) = @_; 
	
	#The following script is to seek all reads from the same RNA.

	###### Extend ######
	my %BBshape=(); 
	$BBshape{$iseq}++; 

	my $needAssemby; 
	#################################################
	my $Xcontrol=1; 
	my $pcontrol=0;

	while($Xcontrol > 0 )
    {
        my %hhain=(); 
		my $swyyy=0; 
	    
		foreach my $xname (keys %BBshape)
        {
		  unless(defined $imatch->{$xname} ) { next; }
		 
		  my $kkk = $imatch->{$xname};
          foreach my $pass (keys %$kkk)
	         {		      
				if(defined $imatch->{$xname}{$pass} )
				 {
					$swyyy++; $pcontrol++;
					$hhain{$pass}++; $needAssemby->{$pass}++; 
					delete $imatch->{$xname}{$pass};
                    my $rxname=REVCOMP($xname); my $rpass=REVCOMP($pass);
					delete $imatch->{$rpass}{$rxname};
				 }
	         }
        }
        undef %BBshape; 
		%BBshape=%hhain; 
		undef %hhain;
	    if($swyyy == 0 ) { $Xcontrol=0; }
	}
  	#############################################
	
	if($pcontrol >= 1)
	{
		my $longFseq = &LocalAssemble($iseq, $needAssemby,$iset,$ikmer,$iminE,$imiAL,$imxAL);
        return $longFseq; 
	}
	else { return 1; }
}


###############################################################################################################################
sub LocalAssemble
{
	my ($yseq,$needAssemby,$binset,$ykmer,$yminE,$ymiAL,$ymxAL) = @_; 
	
	my $zkmer=$ykmer+1;
    

	##############
	#The following is to assemble reads. 
	#All reads would be splitted into smaller fragments of (k-mer+1) nt.
    #############

	undef my $rcount;
	
	my $bbbin; 
    foreach my $xys (keys %$needAssemby)
      {
        my $num=$binset->{$xys}; 
		for(my $i=0; $i <=length($xys); $i++)
		  {
			my $subseq1=substr($xys,$i,$zkmer); my $subseq2=substr($xys,$i,$ykmer);
            if(length($subseq1)==$zkmer) { $bbbin->{$subseq2}{$subseq1}+=$num; $rcount->{$subseq1}+=$num;  }   
		  }
      }
	 
	
	 ###############
	 my $candidateAssembleContig;  #hash to store all assembled sequences.

     ######
	 #the following script is to extend contig.
	 ######

	 my %BBshape=(); 
	 $BBshape{$yseq} = $binset->{$yseq};
	 
	 my $Xcontrol=1;

     while($Xcontrol > 0 )
      {
        my %hhain=(); my $swyyy=0;
        foreach my $xys (keys %BBshape)
         {
           if(length($xys) > $ymxAL) { next; }
		   my $pvalue=$BBshape{$xys};
		   my $startp=length($xys)-$ykmer; my $startSeq=substr($xys,$startp,$ykmer);
		  
		   if( ( defined $bbbin->{$startSeq}) != 1 ) 
			   {  
			     unless(length($xys) >= $ymiAL ) { next; }
				 $candidateAssembleContig->{$xys}=$pvalue; 
			   }

	       elsif(defined $bbbin->{$startSeq} )
	        { 
		       my $kkk = $bbbin->{$startSeq};
               foreach my $pass (keys %$kkk)
	           {		      
				 if($pass=~m/[ATCG]/ && length($pass) == $zkmer )
				 {
					 $swyyy++; 
					 my $newch=substr($pass,$ykmer,1);
					 my $num = $rcount->{$pass};
					 unless($num >= $yminE ) {next; }
					 my $num+=$pvalue; my $aabb="$xys"."$newch"; 
					 $hhain{$aabb}=$num; 
					 delete $bbbin->{$startSeq}{$pass};
				 }
	           }
            }
         }
		 my @tttas = %hhain;
		 if( @tttas > 0  )
			 { 
			   undef %BBshape; 
			   %BBshape=%hhain; 
			 }
		 if( @tttas == 0 || $swyyy == 0) {  $Xcontrol=0 ; }
     }

   my $contigSize; #define the size of the longest contig
   foreach my $yread ( keys %BBshape ) 
	   { 
	      $candidateAssembleContig->{$yread} = $BBshape{$yread}; 
		  $contigSize=length($yread); 
	   }
	 
   my $kookseq= LookOPT($candidateAssembleContig,$contigSize);

   ######
   return $kookseq; 
}

##############################################################################################################
sub  LookOPT
{
   my ($icandidatSet,$icontigSize) = @_;
   
   my $qlength=$icontigSize-25; #range of contig's length.
   
   #####
   #Seeking the longest contig with highest abundant siRNAs. The first 40-nt and last 40-nt of the contig should contain at least 6-nt repeat unit.
   #The segment between two recognized repeat units is supposed to be the candidate sequence.
   #####

   my $optContig=''; my $tagName='';  
   foreach my $yread (sort { $icandidatSet->{$b} <=> $icandidatSet->{$a} } keys %$icandidatSet )
   { 
	 
	 if(length($yread) >= $qlength )   
	 {  
	    $optContig=$yread; my $ttt=length($yread)-40;   #last 40 nt will be used for search matching tag
        unless($optContig=~m/[ATCG]/i) { next; }
		my $xseq=substr($optContig,0,40);
		my $yseq=substr($optContig,$ttt,40);
        
		#####
         for(my $i=6; $i <= 38; $i++)
         {
           for(my $j=0; $j <= length($xseq); $j++)
		   {
              my $atag=substr($xseq,$j,$i);  $atag=~s/\s+//g; 
			  unless(length($atag) == $i ) { next; }
              if($yseq=~m/$atag/) { $tagName=$atag; }
           }
         }
        #####
	    if($tagName=~m/[ATCG]/) { last; }
	 }
   }
   ################################################################## 

  my $finalseq=''; 
  
  if($tagName =~m/[ATCG]/) 
   {
	  $optContig=~s/$tagName/  $tagName  /g; 
      my @seqARRAY=split(/\s+/,$optContig);

      my $start=0; my $end=0;
      for(my $i=0; $i <= $#seqARRAY; $i++)  { if($seqARRAY[$i]=~m/^$tagName/) { $start=$i; last; } }
      for(my $i=$#seqARRAY; $i >= 0; $i--)  { if($seqARRAY[$i]=~m/^$tagName/) { $end=$i; last; } }
      if($start >= 0 && $end >= $start) 
		  { 
		     for(my $i=$start; $i < $end; $i++)  { $finalseq.="$seqARRAY[$i]"; } 
		  }
   }
  
  return $finalseq;
  
}
##############################################################################################################



#################################################################
#reading input file
##

sub readINFILE
{
   my ($ifile,$iminl,$imaxl) = @_;
   
   undef my $set;
   my $count=0;
   
   print "Start reading $ifile\n";

   open(INDATA, $ifile) || die "could not open $ifile\n";
   while(my $line=<INDATA>)
	{  
      unless($line =~ /^[ATCG]/i ) { next; }
      
	  chomp($line);
      $line=~s/\s+//g;
	  unless(length($line) >= $iminl && length($line) <= $imaxl) { next; } 
	  
	  $count++; 
	  print "$count\n" if ( $count%10000 == 0); 

	  ####
	  #the complemenatry sequence of each read is generated automatically.
      ####

	  my $orig = uc($line);  my $rorig=&REVCOMP($orig); 
	  
	  $set->{$orig}++; $set->{$rorig}++;
   }
   close INDATA;
  
   print "Total reads: $count\n";
   return $set;
}
###############################################################################################################################

sub filterINPUTset
{
   
   my ($yinputset, $ykmerST) = @_;

   my $filterlist;
   my $totalread=0;

   print "\nStart removing non-overlapping and terminal reads:\n";
   while((keys %$yinputset) > 0)
    {
      my $yinputbin = &creatHASH($yinputset,$ykmerST);
	  
	  my $newTagNum=0;
	  
	  undef my $leftset;
      foreach my $tag (keys %$yinputset) 
       {  
		 unless($tag=~m/[ATCG]/) { next; }
	     if(defined $leftset->{$tag} )   { next; }

		 my $rtag= &REVCOMP($tag);
		 if( &seekMATCHP($tag, $yinputbin, $ykmerST)  || &seekMATCHP($rtag, $yinputbin, $ykmerST) )
	     {
            $newTagNum++; $totalread++;
			print "." if ($totalread%1000==0);
		    delete $yinputset->{$tag};  delete $yinputset->{$rtag};
            $leftset->{$tag}++;  $leftset->{$rtag}++;
	     }
       }
	  keepFILTERdata($yinputset,$ykmerST);

  	  if( $newTagNum == 0 ) { $filterlist = $yinputset; print "\n"; last;  }
    } 
   
	return $filterlist;
}

####################################################################################################################
#the following script is to classfy small reads into non-overlapping/terminal reads and internal overalpping reads.
##

sub seekMATCHP 
{
	my ($iseq, $ibin, $ikmer) = @_;
    
	my $isize=length($iseq); 
	my $ipair=1;

	for (my $index=($isize-$ikmer); $index >=0; $index--) 
	 {  
        my $subsubseq  = substr($iseq,$index,$ikmer);    $subsubseq=~s/\s+//g;
		unless(length($subsubseq) == $ikmer ) { next; }

		my $subsubseq2 = substr($iseq,$index,$isize);    $subsubseq2=~s/\s+//g; 
        unless(length($subsubseq2) >= $ikmer ) { next; }

		unless( defined $ibin->{$subsubseq}) { next; }
		my $subset = $ibin->{$subsubseq}; 
		foreach my $pass ( keys %$subset )
	       {		      
				if($pass=~m/^$subsubseq2/ && length($pass) > length($subsubseq2)  )
				{ $ipair=0; goto H2;}
		   }
	 }
	 H2:
	 return $ipair;
}

####################################################################################################################3
sub creatHASH 
{
	my ($iset, $ikmer) = @_;
	
	my $ibin;

    foreach my $xys (keys %$iset)
    {
      if($xys=~m/^[ATCG]/)
	  {
		my $sf=substr($xys,0,$ikmer); $sf=~s/\s+//g;
		if(length($sf)==$ikmer && length($sf) < length($xys) ) {  $ibin->{$sf}{$xys} = $iset->{$xys}; }
	  }
    }
    return $ibin;
}
####################################################################################################################3
sub REVCOMP 
{
	$_ = shift;
	$_ = uc();
	tr/ATGC/TACG/;
	return (reverse());
}

#############################################################################

sub keepFILTERdata
{
  my ($yinputset,$ikmer) = @_;
  
  open(PUT, ">$inputfile.$ikmer");

  foreach my $tag (keys %$yinputset) 
  {  
    unless($tag=~m/[ATCG]/) { next; }
    my $tt = $yinputset->{$tag};
    print PUT ">$tag";
    print PUT "_$tt\n$tag\n";
  }
  close(PUT);
  
  return 1;
}
#############################################################################
sub seekMATCHB 
{
	my ($iseq, $ibin, $ikmer) = @_;
    
	my $isize=length($iseq); 
	my $ipair;

	for (my $index=0; $index <= ($isize-$ikmer); $index++) 
	 {  
        my $subsubseq  = substr($iseq,$index,$ikmer);    $subsubseq=~s/\s+//g;
		my $subsubseq2 = substr($iseq,$index,$isize);    $subsubseq2=~s/\s+//g; 

        if(length($subsubseq) == $ikmer && length($subsubseq2) >= $ikmer ) 
	    { 
		   unless( defined $ibin->{$subsubseq}) { next; }
		   my $subset = $ibin->{$subsubseq}; 
		   foreach my $pass ( keys %$subset )
	          {		      
				if($pass=~m/^$subsubseq2/ && length($pass) > length($subsubseq2)  )
				{ my $xxx="$iseq".','."$pass";  $ipair->{$xxx}++; }
		      }
		}
	 }
	 return $ipair;
}
#############################################################################
